@@ -36,7 +36,6 @@ import Data.ByteString.Builder (byteString, Builder)
 import Data.ByteString.Builder.Extra (flush)
 import qualified Data.CaseInsensitive as CI
 import Data.Function (on)
-import Data.Hashable (hash)
 import Data.List (deleteBy)
 import Data.Maybe
 #if MIN_VERSION_base(4,5,0)
@@ -57,9 +56,6 @@ import qualified Network.HTTP.Types.Header as H
 import Network.Wai
 import Network.Wai.Handler.Warp.Buffer (toBuilderBuffer)
 import qualified Network.Wai.Handler.Warp.Date as D
-#ifndef WINDOWS
-import qualified Network.Wai.Handler.Warp.FdCache as F
-#endif
 import Network.Wai.Handler.Warp.File
 import Network.Wai.Handler.Warp.Header
 import Network.Wai.Handler.Warp.IO (toBufIOWith)
@@ -148,14 +144,14 @@ sendResponse settings conn ii req reqidxhdr src response = do
         -- and status, the response to HEAD is processed here.
         --
         -- See definition of rsp below for proper body stripping.
-        (ms, mlen) <- sendRsp conn ii req ver s hs rsp
+        (ms, mlen) <- sendRsp conn ii ver s hs rsp
         case ms of
             Nothing         -> return ()
             Just realStatus -> logger req realStatus mlen
         T.tickle th
         return ret
       else do
-        _ <- sendRsp conn ii req ver s hs RspNoBody
+        _ <- sendRsp conn ii ver s hs RspNoBody
         logger req s Nothing
         T.tickle th
         return isPersist
@@ -167,8 +163,8 @@ sendResponse settings conn ii req reqidxhdr src response = do
     hs0 = sanitizeHeaders $ responseHeaders response
     rspidxhdr = indexResponseHeader hs0
     th = threadHandle ii
-    dc = dateCacher ii
-    addServerAndDate = addDate dc rspidxhdr . addServer defServer rspidxhdr
+    getdate = getDate ii
+    addServerAndDate = addDate getdate rspidxhdr . addServer defServer rspidxhdr
     (isPersist,isChunked0) = infoFromRequest req reqidxhdr
     isChunked = not isHead && isChunked0
     (isKeepAlive, needsChunked) = infoFromResponse rspidxhdr (isPersist,isChunked)
@@ -225,7 +221,6 @@ data Rsp = RspNoBody
 
 sendRsp :: Connection
         -> InternalInfo
-        -> Request
         -> H.HttpVersion
         -> H.Status
         -> H.ResponseHeaders
@@ -234,7 +229,7 @@ sendRsp :: Connection
 
 ----------------------------------------------------------------
 
-sendRsp conn _ _ ver s hs RspNoBody = do
+sendRsp conn _ ver s hs RspNoBody = do
     -- Not adding Content-Length.
     -- User agents treats it as Content-Length: 0.
     composeHeader ver s hs >>= connSendAll conn
@@ -242,7 +237,7 @@ sendRsp conn _ _ ver s hs RspNoBody = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ _ ver s hs (RspBuilder body needsChunked) = do
+sendRsp conn _ ver s hs (RspBuilder body needsChunked) = do
     header <- composeHeaderBuilder ver s hs needsChunked
     let hdrBdy
          | needsChunked = header <> chunkedTransferEncoding body
@@ -255,7 +250,7 @@ sendRsp conn _ _ ver s hs (RspBuilder body needsChunked) = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ _ ver s hs (RspStream streamingBody needsChunked th) = do
+sendRsp conn _ ver s hs (RspStream streamingBody needsChunked th) = do
     header <- composeHeaderBuilder ver s hs needsChunked
     (recv, finish) <- newBlazeRecv $ reuseBufferStrategy
                     $ toBuilderBuffer (connWriteBuffer conn) (connBufferSize conn)
@@ -279,7 +274,7 @@ sendRsp conn _ _ ver s hs (RspStream streamingBody needsChunked th) = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ _ _ _ _ (RspRaw withApp src tickle) = do
+sendRsp conn _ _ _ _ (RspRaw withApp src tickle) = do
     withApp recv send
     return (Nothing, Nothing)
   where
@@ -293,10 +288,9 @@ sendRsp conn _ _ _ _ _ (RspRaw withApp src tickle) = do
 
 -- Sophisticated WAI applications.
 -- We respect s0. s0 MUST be a proper value.
-sendRsp conn ii req ver s0 hs0 (RspFile path (Just part) _ isHead hook) =
-    sendRspFile2XX conn ii req ver s0 hs h path beg len isHead hook
+sendRsp conn ii ver s0 hs0 (RspFile path (Just part) _ isHead hook) =
+    sendRspFile2XX conn ii ver s0 hs path beg len isHead hook
   where
-    h = hash $ rawPathInfo req
     beg = filePartOffset part
     len = filePartByteCount part
     hs = addContentHeadersForFilePart hs0 part
@@ -305,60 +299,47 @@ sendRsp conn ii req ver s0 hs0 (RspFile path (Just part) _ isHead hook) =
 
 -- Simple WAI applications.
 -- Status is ignored
-sendRsp conn ii req ver _ hs0 (RspFile path Nothing idxhdr isHead hook) = do
-    let h = hash $ rawPathInfo req
-    efinfo <- E.try $ fileInfo' ii h path
+sendRsp conn ii ver _ hs0 (RspFile path Nothing idxhdr isHead hook) = do
+    efinfo <- E.try $ getFileInfo ii path
     case efinfo of
         Left (_ex :: E.IOException) ->
 #ifdef WARP_DEBUG
           print _ex >>
 #endif
-          sendRspFile404 conn ii req ver hs0
+          sendRspFile404 conn ii ver hs0
         Right finfo -> case conditionalRequest finfo hs0 idxhdr of
-          WithoutBody s         -> sendRsp conn ii req ver s hs0 RspNoBody
-          WithBody s hs beg len -> sendRspFile2XX conn ii req ver s hs h path beg len isHead hook
+          WithoutBody s         -> sendRsp conn ii ver s hs0 RspNoBody
+          WithBody s hs beg len -> sendRspFile2XX conn ii ver s hs path beg len isHead hook
 
 ----------------------------------------------------------------
 
 sendRspFile2XX :: Connection
                -> InternalInfo
-               -> Request
                -> H.HttpVersion
                -> H.Status
                -> H.ResponseHeaders
-               -> Hash
                -> FilePath
                -> Integer
                -> Integer
                -> Bool
                -> IO ()
                -> IO (Maybe H.Status, Maybe Integer)
-sendRspFile2XX conn ii req ver s hs h path beg len isHead hook
-  | isHead = sendRsp conn ii req ver s hs RspNoBody
+sendRspFile2XX conn ii ver s hs path beg len isHead hook
+  | isHead = sendRsp conn ii ver s hs RspNoBody
   | otherwise = do
       lheader <- composeHeader ver s hs
-#ifdef WINDOWS
-      let fid = FileId path Nothing
-          hook' = hook
-#else
-      (mfd, hook') <- case fdCacher ii of
-         -- settingsFdCacheDuration is 0
-         Nothing  -> return (Nothing, hook)
-         Just fdc -> do
-            (fd, fresher) <- F.getFd' fdc h path
-            return (Just fd, hook >> fresher)
+      (mfd, fresher) <- getFd ii path
       let fid = FileId path mfd
-#endif
+          hook' = hook >> fresher
       connSendFile conn fid beg len hook' [lheader]
       return (Just s, Just len)
 
 sendRspFile404 :: Connection
                -> InternalInfo
-               -> Request
                -> H.HttpVersion
                -> H.ResponseHeaders
                -> IO (Maybe H.Status, Maybe Integer)
-sendRspFile404 conn ii req ver hs0 = sendRsp conn ii req ver s hs (RspBuilder body True)
+sendRspFile404 conn ii ver hs0 = sendRsp conn ii ver s hs (RspBuilder body True)
   where
     s = H.notFound404
     hs =  replaceHeader H.hContentType "text/plain; charset=utf-8" hs0
@@ -435,10 +416,10 @@ addTransferEncoding hdrs = (H.hTransferEncoding, "chunked") : hdrs
 addTransferEncoding hdrs = ("transfer-encoding", "chunked") : hdrs
 #endif
 
-addDate :: D.DateCache -> IndexedHeader -> H.ResponseHeaders -> IO H.ResponseHeaders
-addDate dc rspidxhdr hdrs = case rspidxhdr ! fromEnum ResDate of
+addDate :: IO D.GMTDate -> IndexedHeader -> H.ResponseHeaders -> IO H.ResponseHeaders
+addDate getdate rspidxhdr hdrs = case rspidxhdr ! fromEnum ResDate of
     Nothing -> do
-        gmtdate <- D.getDate dc
+        gmtdate <- getdate
         return $ (H.hDate, gmtdate) : hdrs
     Just _ -> return hdrs
 
